#
# Copyright 2010 Andrey Churin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import sys
import re
import string
import types

from closurebuilder.util import Error

__all__ = ['read_manifest', 'manifest_directive']

def manifest_directive(phase=0):
    def decorator(fn):
        fn.directive = True
        if not isinstance(phase, (list, tuple)):
            p = (phase,)
        else:
            p = phase
        fn.phase = p
        return fn
    if isinstance(phase, types.FunctionType):
        fn = phase
        phase = 0
        return decorator(fn)
    return decorator

def get_indent(s):
    "Gets string indentation"
    return len(s) - len(s.lstrip())

def read_lines(filename):
    "Reads lines from file"

    if not os.path.isfile(filename):
        raise Error('File "%s" not found'%filename)
    lines = []
    f = open(filename, 'rt')
    for num, line in enumerate(f.readlines()):
        line = line.rstrip()
        stripped = line.lstrip()
        if stripped and not stripped.startswith('#'):
            lines.append((num + 1, line, None, None))
    f.close()
    return lines

def get_block(lines, indent):
    result = []
    if not lines:
        raise Error('Expected an indented block')
    base_indent = get_indent(lines[0][1])
    if base_indent < indent:
        raise Error('Expected an indented block')
    while lines and base_indent <= get_indent(lines[0][1]):
        result.append(lines.pop(0))
    return Block(result, base_indent)

def get_args(line, processor):
    for i in dir(processor):
        meth = getattr(processor, i)
        if not hasattr(meth, 'directive'):
            continue
        match = meth.__doc__.replace(' ', r'\s+'). \
                    replace('%', '([^"\s]+|"[^"]*")')
        match = re.match('^' + match + '$', line)
        if match:
            result = []
            for i in match.groups():
                if i.startswith('"') and i.endswith('"'):
                    i = i[1:-1]
                result.append(i)
            return meth, result

    raise Error('Invalid directive', 0)

def read_manifest(filename, block=None, processor=None):
    "Reads manifest file"

    if not block:
        lines = read_lines(filename)
        indent = 0
    else:
        lines = block._lines
        indent = block._indent

    phase = 0
    next_phase_lines = lines
    while next_phase_lines:
        next_phase_lines = []
        while lines:
            num, line, meth, args = lines.pop(0)
            try:
                if not args:
                    if get_indent(line) != indent:
                        raise Error('Unexpected indent', 0)

                    if line.endswith(':'):
                        line = line[:-1]
                        block = True
                    else:
                        block = False

                    meth, args = get_args(line.strip(), processor)
                    func = meth.im_func
                    varnames = func.func_code.co_varnames[2:\
                                                func.func_code.co_argcount]
                    if func.func_defaults:
                        defaults = len(func.func_defaults)
                    else:
                        defaults = 0

                    if not block:
                        if not defaults and varnames and varnames[-1] == 'block_':
                            raise Error('Expected an indented block', 0)
                    elif block:
                        if not varnames or varnames[-1] != 'block_':
                            raise Error('Unexpected indented block', 0)
                        args.append(get_block(lines, indent))

                    if not((len(varnames) - defaults) <= len(args) \
                                                           <= len(varnames)):
                        raise Error('Invalid arguments count', 0)
                if not phase in meth.phase:
                    next_phase_lines.append((num, line, meth, args))
                else:
                    meth(*([Opt(filename, num, phase)] + args))
                    if max(meth.phase) > phase:
                        next_phase_lines.append((num, line, meth, args))

            except Error, e:
                raise e.traceback('File "%s", line %s\n  %s'%(filename, num,
                                                                        line))
        lines = next_phase_lines
        phase+=1

    if hasattr(processor, 'check_integrity'):
        try:
            processor.check_integrity()
        except Error, e:
            raise e.traceback('File "%s"' % filename)

    return processor

class Block(object):
    def __init__(self, lines, indent):
        self._lines = lines
        self._indent = indent

class Opt(object):
    def __init__(self, filename, lineno, phase):
        self.filename = filename
        self.lineno = lineno
        self.phase = phase
