#
# Copyright 2010 Andrey Churin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import re

from closurebuilder.metadata import Metadata
from closurebuilder.manifest import read_manifest, manifest_directive, get_indent
from closurebuilder.util import validate_namespace, Error

NAME_REGEX = re.compile(r'^[_a-zA-Z0-9]+$')

class ModuleBase(object):

    def __init__(self):
        self.entry_point = None
        self._write_to = None
        self._provides = None
        self._include = []
        self._sources = None

    def get_module_name(self):
        "Gets module name"
        return self._write_to or self.entry_point.replace('.', '_')

    def get_sources(self):
        if self._sources is None:
            meta = Metadata.get_instance()
            self._sources = meta.sources.get_dependencies(self.entry_point)
        return self._sources

    def get_all_sources(self):
        sources = set(self._include)
        sources.update(self.get_sources())
        return sources

    def get_provides(self):
        if self._provides is not None:
            return self._provides
        self._provides = set()
        meta = Metadata.get_instance()
        for i in self.get_sorted_sources():
            self._provides.update(i.provides)
        return self._provides

    def get_resources(self):
        files = []
        for i in self.get_sorted_sources():
            path = os.path.dirname(i.get_path())
            for res in i.resources:
                res = os.path.join(path, res)
                if res in files:
                    raise Error('The resource file "%s" was added '
                                'more then once'%res)
                files.append(res)
        return files

    def add_include(self, include):
        if not i in self._include:
            self._include.append(i)

    @manifest_directive
    def write_to(self, opt, filename):
        "write to %"
        self._write_to = filename


class Application(ModuleBase):
    "Application"

    def __init__(self, path):
        super(Application, self).__init__()
        self.path = path
        self.modules = []
        self.target = None
        self._write_to = None
        self._write_css_to = None
        self._compiler_settings = []

    def get_compiler_settings(self):
        return self._compiler_settings

    def get_dep_paths(self):
        return [[self]]

    def get_css_filename(self):
        return self._write_css_to

    def get_sorted_sources(self):
        return self.get_sources()

    @manifest_directive(phase=(1,2))
    def compile_application(self, opt, name):
        "compile application %"
        if opt.phase == 1:
            if self.entry_point:
                raise Error('Directive is not allowed here', 0)
            validate_namespace(name, 20)
            if not Metadata.get_instance().sources.find_sources(provide=name):
                raise Error('Namespace "%s" not found'%name, 20)
            self.entry_point = name
        if self.modules:
            source = Metadata.get_instance().sources.find_sources(provide=name)
            source.requires.add('goog.module.ModuleManager')
            source.requires.add('goog.module.ModuleLoader')


    @manifest_directive(phase=(1,2))
    def compile_module(self, opt, name, block_ = []):
        "compile module %"
        if opt.phase == 1:
            validate_namespace(name, 15)
            for i in self.modules:
              if i.entry_point == name:
                raise Error('Namespace "%s" is used already'%name, 15)

            meta = Metadata.get_instance()
            requires = meta.sources.find_sources(require=name)
            if requires:
                raise Error('Namespace "%s" is required in these sources:\n%s'%
                                (name, '\n'.join(requires)), 15)
            if meta.sources.find_sources(provide=name):
                module = Module(self, name)
                self.modules.append(module)
            else:
                raise Error('Namespace "%s" not found'%name, 15)
        if opt.phase == 2:
            for i in self.modules:
                if i.entry_point == name:
                    read_manifest(opt.filename, block_, i)
                    break

    @manifest_directive
    def enable_debug(self, opt):
        "enable debug"
        self._compiler_settings.append('--debug')

    @manifest_directive
    def define(self, opt, value):
        "define %"
        self._compiler_settings += '--define', value

    @manifest_directive
    def add(self, opt, value):
        "add extern %"
        path = os.path.join(self.path, value)
        if not os.path.isfile(path):
            raise Error('File "%s" does not exists'%path, 11)
        self._compiler_settings += '--externs', os.path.abspath(path)

    @manifest_directive
    def write_css_to(self, opt, filename):
        "write css to %"
        self._write_css_to = filename

    @manifest_directive
    def scan(self, opt, path):
        "scan %"
        path = os.path.join(self.path, path)
        Metadata.get_instance().sources.scan_(opt, path)

    @manifest_directive
    def scan_loose(self, opt, path):
        "scan loose %"
        path = os.path.join(self.path, path)
        Metadata.get_instance().sources.scan_loose_(opt, path)

    @manifest_directive
    def set_target(self, opt, folder):
        "set target %"
        self.target = os.path.join(Metadata.get_instance().closure_dir, folder)

    def check_integrity(self):
        if not self.entry_point:
            raise Error('Directive "compile application" expected')
        if not self.target:
            raise Error('Directive "set target" expected')


class Module(ModuleBase):
    "Module metadata"

    def __init__(self, app, entry_point):
        super(Module, self).__init__()
        self.app = app
        self.entry_point = entry_point
        self.uri = None
        self._depends_on = []
        self._metadata = None

    @manifest_directive
    def set_uri(self, opt, value):
        "set uri %"
        self.uri = value

    @manifest_directive
    def set_metadata(self, opt, block_):
        "set metadata"
        self._metadata = ModuleMetadata()
        read_manifest(opt.filename, block_, self._metadata)

    @manifest_directive
    def depends_on(self, opt, module):
        "depends on %"
        for i in self.app.modules:
            if i.entry_point == module:
                self._depends_on.append(i)
                break
        else:
            raise Error('Module "%s" is not defined'%module, 11)
        def check_circular(module, circular):
            for i in module._depends_on:
                if i in circular:
                    raise Error('Encountered circular module dependency', 11)
                if i != self.app:
                    check_circular(i, circular + [i])
        check_circular(self, [self])

    def check_integrity(self):
        if not self._depends_on:
            self._depends_on.append(self.app)
        if not self.uri:
            raise Error('Module URI must be specified')

    def get_meta(self):
        if self._metadata:
            return str(self._metadata)
        return None

    def get_all_sources(self):
        sources = super(Module, self).get_all_sources()
        for i in self._depends_on:
            sources.update(i.get_all_sources())
        return sources

    def get_sorted_sources(self):
        all_sources = set()
        for i in self._depends_on:
            all_sources.update(i.get_all_sources())

        return filter(lambda x: x not in all_sources,
                            self.get_sources() + self._include)

    def get_depends_on(self):
        return self._depends_on

    def get_dep_paths(self):
        result = []
        for i in self._depends_on:
            result.extend(i.get_dep_paths())
        for i in result:
            i.append(self)
        return result


class ModuleMetadata(object):
    "Javascript module metadata"

    def __init__(self):
        self.meta = {}

    @manifest_directive
    def set(self, opt, name, value):
        "set % %"
        if not NAME_REGEX.match(name):
            raise Error('Invalid metadata field name', 4)
        self.meta[name] = value

    def __str__(self):
        return '{%s}'%','.join(map(lambda x: '%s:"%s"'%(x, self.meta[x]),
                                        self.meta))
